/***************************************************************************
 *   Copyright (C) 2008 by C J Plooy                                       *
 *   cornwarecjp@lavabit.com                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Library General Public License as       *
 *   published by the Free Software Foundation; either version 2 of the    *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Library General Public     *
 *   License along with this program; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <cstdio>

#include "colors.h"

COLORREF OrbitMFDPlanetColor      = MFD_GREY;
COLORREF OrbitMFDVesselOrbitColor = MFD_LIGHT_GREEN;
COLORREF OrbitMFDTargetOrbitColor = MFD_YELLOW;
COLORREF OrbitMFDGreenAlertColor  = MFD_LIGHT_GREEN;
COLORREF OrbitMFDYellowAlertColor = MFD_YELLOW;
COLORREF OrbitMFDRedAlertColor    = MFD_RED;

#define MAX_LINE_LEN 255

char *readNextLine(FILE *fp)
{
	static char buffer[MAX_LINE_LEN+1];
	char *ptr = buffer;

	//First character
	{
		int c = fgetc(fp);
		if(c == EOF) return NULL; //End Of File
		if(c == '\n' || c == '\r') return "";
		*ptr = c;
		ptr++;
	}

	//Next characters
	while(ptr - buffer < MAX_LINE_LEN)
	{
		int c = fgetc(fp);
		if(c == EOF || c == '\n' || c == '\r') break;

		*ptr = c;
		ptr++;
	}

	*ptr = '\0'; //make NUL-terminated

	return buffer;
}

inline bool isspace(char c)
	{return c == ' ' || c == '\t';}

inline bool iscomment(char c)
	{return c == ';' || c == '#';}

void readNextValue(FILE *fp, char **name, char **value)
{
	while(true)
	{
		char *line = readNextLine(fp);
		if(line == NULL)
		{
			*name = NULL;
			*value = NULL;
			return;
		}

		if(*line == '\0') continue; //empty line

		//Skip whitespace
		while(*line != '\0')
		{
			if(!isspace(*line)) break;
			line++;
		}

		if(*line == '\0') continue; //whitespace-only line

		//This should be the start of the name:
		*name = line;

		//Search end of name, and check for comments
		bool comment = false;
		bool equal   = false;
		while(*line != '\0')
		{
			if(iscomment(*line)) comment = true;
			if(*line == '=') equal = true;
			if(comment || equal || isspace(*line)) break;
			line++;
		}

		if(*line == '\0') continue; //name only
		if(comment) continue; //name and comment only

		*line = '\0'; //end of name
		line++;

		//Search for equal sign if not already found
		if(!equal)
		{
			while(*line != '\0')
			{
				if(*line == '=') equal = true;
				if(!isspace(*line)) break;
				line++;
			}

			if(!equal) continue; //name, not followed directly by equal

			line++; //skip equal sign
		}

		//Skip whitespace
		while(*line != '\0')
		{
			if(!isspace(*line)) break;
			line++;
		}

		//This should be the start of the value:
		*value = line;

		//Check for comment:
		if(iscomment(*line)) *line = '\0'; //make empty string

		if(*line == '\0') return; //empty string value

		//Trim right, and check for comments:
		char *lastNonSpace = line;
		while(*line != '\0' && !iscomment(*line))
		{
			if(!isspace(*line)) lastNonSpace = line;
			line++;
		}

		*(lastNonSpace+1) = '\0'; //end of value
		return;
	}
}

COLORREF str2color(char *s)
{
	unsigned int num;
	if(sscanf(s, "%x", &num) != 1) return RGB(0xFF,0xFF,0xFF);

	unsigned char b = num & 0xFF;
	unsigned char g = (num >> 8) & 0xFF;
	unsigned char r = (num >> 16) & 0xFF;

	return RGB(r,g,b);
}

void loadColorConfiguration()
{
	FILE *fp = fopen("FreeOrbitMFD.cfg", "rb");
	if(fp == NULL) return; //keeps compiled-in defaults

	while(true)
	{
		char *name, *value;
		readNextValue(fp, &name, &value);
		if(name == NULL || value == NULL) break;

		if(strcmp(name, "PlanetColor") == 0)
		{
			OrbitMFDPlanetColor = str2color(value);
		}
		else if(strcmp(name, "VesselOrbitColor") == 0)
		{
			OrbitMFDVesselOrbitColor = str2color(value);
		}
		else if(strcmp(name, "TargetOrbitColor") == 0)
		{
			OrbitMFDTargetOrbitColor = str2color(value);
		}
		else if(strcmp(name, "GreenAlertColor") == 0)
		{
			OrbitMFDGreenAlertColor = str2color(value);
		}
		else if(strcmp(name, "YellowAlertColor") == 0)
		{
			OrbitMFDYellowAlertColor = str2color(value);
		}
		else if(strcmp(name, "RedAlertColor") == 0)
		{
			OrbitMFDRedAlertColor = str2color(value);
		}
	}

	fclose(fp);
}

