/***************************************************************************
 *   Copyright (C) 2008 by C J Plooy                                       *
 *   cornwarecjp@lavabit.com                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Library General Public License as       *
 *   published by the Free Software Foundation; either version 2 of the    *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Library General Public     *
 *   License along with this program; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#define STRICT
#define ORBITER_MODULE

#include <cstdio>
#include <cstdlib>

#include <algorithm> //for std::swap
#include <vector>

#include <windows.h>

#include "orbitersdk.h"

#include "orbitersdk_ext_cjp.h"
#include "kost.h"

#include "orbitmfd.h"

#include "colors.h"

unsigned int numDecimals(double value)
{
	value = fabs(value);

	unsigned int ret = 0;
	while(value < 1000.0 && ret < 3)
	{
		value *= 10.0;
		ret++;
	}
	return ret;
}

void formatValue(char *str, double value)
{
	double a = fabs(value);

	if(a < 1000.0)      //1km
	{
		sprintf(str, "% 0.*f", numDecimals(value), value);
	}
	else if(a < 1e6)    //1000km
	{
		value /= 1000.0;
		sprintf(str, "% 0.*fk", numDecimals(value), value);
	}
	else if(a < 1e9)    //1 million km
	{
		value /= 1e6;
		sprintf(str, "% 0.*fM", numDecimals(value), value);
	}
	else if(a < 10e3 * KOST_ASTRONOMICAL_UNIT) //10000AU
	{
		value /= KOST_ASTRONOMICAL_UNIT;
		sprintf(str, "% 0.*fAU", numDecimals(value), value);
	}
	else if(a < 1000.0 * KOST_PARSEC)          //1000pc
	{
		value /= KOST_PARSEC;
		sprintf(str, "% 0.*fpc", numDecimals(value), value);
	}
	else if(a < 1e6 * KOST_PARSEC)             //1Mpc
	{
		value /= (1000.0*KOST_PARSEC);
		sprintf(str, "% 0.*fkpc", numDecimals(value), value);
	}
	else if(a < 1e9 * KOST_PARSEC)             //1Gpc
	{
		value /= (1e6 * KOST_PARSEC);
		sprintf(str, "% 0.*fMpc", numDecimals(value), value);
	}
	else
	{
		value /= (1e9 * KOST_PARSEC);
		sprintf(str, "% 0.*fGpc", numDecimals(value), value);
	}
}

bool refMenuResponse(void *menuPtr, unsigned int menuIndex, char *name, void *userPtr)
{
	COrbitMFD *mfd = (COrbitMFD *)userPtr;

	OBJHANDLE newRef = oapiGetGbodyByName(name);

	if(newRef == NULL) return false; //Reject non-existent objects

	//Reject non-celestial body objects:
	int objType = oapiGetObjectType(newRef);
	if(objType != OBJTP_STAR && objType != OBJTP_PLANET) return false;

	//Assign reference:
	mfd->m_Reference = newRef;
	mfd->InvalidateDisplay();

	return true; //accept selection
}

bool tgtMenuResponse(void *menuPtr, unsigned int menuIndex, char *name, void *userPtr)
{
	COrbitMFD *mfd = (COrbitMFD *)userPtr;

	OBJHANDLE newTgt = oapiGetObjectByName(name);

	if(newTgt == NULL) return false; //Reject non-existent objects

	//Assign target:
	mfd->m_Target = newTgt;
	mfd->InvalidateDisplay();

	return true; //accept selection
}

int FreeOrbitMFDMode = 0;

/*
Static storage of MFD state:
*/
struct
{
	OBJHANDLE m_Reference;
	OBJHANDLE m_Target;
	bool m_ShowAlt;
	COrbitMFD::eFrame      m_Frame;
	COrbitMFD::eProjection m_Projection;

	COrbitMFD::eMode       m_Mode;

} _FreeOrbitMFDState;

DLLCLBK void InitModule (HINSTANCE hModule)
{
	static char *name = "Free Orbit MFD";
	MFDMODESPEC spec;
	spec.name    = name;
	spec.key     = OAPI_KEY_F; // No longer used according to Orbiter docs
	spec.msgproc = COrbitMFD::MsgProc;

	FreeOrbitMFDMode = oapiRegisterMFDMode(spec);

	//Initial "un-initialized" state:
	_FreeOrbitMFDState.m_Reference = NULL;
	_FreeOrbitMFDState.m_Target    = NULL;

	//Load user-defined colors:
	loadColorConfiguration();
}

DLLCLBK void ExitModule (HINSTANCE hModule)
{
	oapiUnregisterMFDMode(FreeOrbitMFDMode);
}

COrbitMFD::COrbitMFD(DWORD w, DWORD h, VESSEL *vessel) : MFD (w, h, vessel)
{
	m_Width = w;
	m_Height = h;
	m_Vessel = vessel;

	//Allocate shape points:
	m_VesselData.shape.numPoints = 100;
	m_TargetData.shape.numPoints = 100;

	m_VesselData.shape.points = new kostVector3[m_VesselData.shape.numPoints];
	m_TargetData.shape.points = new kostVector3[m_TargetData.shape.numPoints];

	//Initial mode:
	m_ShowAlt    = false;
	m_Frame      = eFEcliptic;
	m_Projection = ePFrame;
	m_Target     = NULL;
	m_Mode       = eMBoth;

	//Initialize reference body:
	autoReference();
}

COrbitMFD::~COrbitMFD()
{
	//De-allocate shape points:
	delete [] m_VesselData.shape.points;
	delete [] m_TargetData.shape.points;
}

int COrbitMFD::MsgProc(UINT msg, UINT mfd, WPARAM wparam, LPARAM lparam)
{
	switch (msg)
	{
	case OAPI_MSG_MFD_OPENED:
		return (int)(new COrbitMFD(LOWORD(wparam), HIWORD(wparam), (VESSEL*)lparam));
	}
	return 0;
}

void COrbitMFD::WriteStatus(FILEHANDLE scn) const
{
	//oapiWriteScenario_xxx

	char name[256];

	switch(m_Projection)
	{
	case ePMax:
	case ePFrame:
		oapiWriteScenario_string(scn, "PROJ", "Frame");
		break;
	case ePShip:
		oapiWriteScenario_string(scn, "PROJ", "Ship");
		break;
	case ePTarget:
		oapiWriteScenario_string(scn, "PROJ", "Target");
		break;
	}

	switch(m_Frame)
	{
	case eFMax:
	case eFEcliptic:
		oapiWriteScenario_string(scn, "FRAME", "Ecliptic");
		break;
	case eFEquator:
		oapiWriteScenario_string(scn, "FRAME", "Equator");
		break;
	}

	if(m_ShowAlt)
		oapiWriteScenario_string(scn, "ALT", "");

	oapiGetObjectName(m_Reference, name, 255);
	oapiWriteScenario_string(scn, "REF", name);

	if(m_Target != NULL)
	{
		oapiGetObjectName(m_Target, name, 255);
		oapiWriteScenario_string(scn, "TARGET", name);
	}
}

void COrbitMFD::ReadStatus(FILEHANDLE scn)
{
	//oapiReadScenario_nextline

	//Default:
	m_ShowAlt = false;
	m_Frame = eFEcliptic;
	m_Projection = ePFrame;
	m_Target = NULL;

	char *line = NULL;
	while(oapiReadScenario_nextline(scn, line))
	{
		unsigned int len = strlen(line);

		if(len > 5 && strncmp(line, "PROJ ", 5) == 0)
		{
			char *projname = line+5;
			len -= 5;

			if(len == 5 && strcmp(projname, "Frame") == 0)
				{m_Projection = ePFrame; continue;}

			if(len == 4 && strcmp(projname, "Ship") == 0)
				{m_Projection = ePShip; continue;}

			if(len == 6 && strcmp(projname, "Target") == 0)
				{m_Projection = ePTarget; continue;}

			continue;
		}

		if(len > 6 && strncmp(line, "FRAME ", 6) == 0)
		{
			char *frmname = line+6;
			len -= 6;

			if(len == 8 && strcmp(frmname, "Ecliptic") == 0)
				{m_Frame = eFEcliptic; continue;}

			if(len == 7 && strcmp(frmname, "Equator") == 0)
				{m_Frame = eFEquator; continue;}

			continue;
		}

		if(len == 3 && strcmp(line, "ALT") == 0)
		{
			m_ShowAlt = true;
			continue;
		}

		if(len > 4 && strncmp(line, "REF ", 4) == 0)
		{
			char *refname = line+4;
			OBJHANDLE newRef = 	oapiGetGbodyByName(refname);

			if(newRef == NULL) continue; //Reject non-existent objects

			//Reject non-celestial body objects:
			int objType = oapiGetObjectType(newRef);
			if(objType != OBJTP_STAR && objType != OBJTP_PLANET) continue;

			//Assign reference:
			m_Reference = newRef;
		}

		if(len > 7 && strncmp(line, "TARGET ", 7) == 0)
		{
			char *tgtname = line+7;
			OBJHANDLE newTgt = oapiGetObjectByName(tgtname);

			if(newTgt == NULL) continue; //Reject non-existent objects

			//Assign target:
			m_Target = newTgt;
		}
	}

	//Don't trust the sanity of a scenario file:
	if(m_Projection == ePTarget && m_Target == NULL) m_Projection = ePFrame;
}

void COrbitMFD::StoreStatus() const
{
	_FreeOrbitMFDState.m_Reference  = m_Reference;
	_FreeOrbitMFDState.m_Target     = m_Target;
	_FreeOrbitMFDState.m_ShowAlt    = m_ShowAlt;
	_FreeOrbitMFDState.m_Frame      = m_Frame;
	_FreeOrbitMFDState.m_Projection = m_Projection;

	_FreeOrbitMFDState.m_Mode       = m_Mode;
}

void COrbitMFD::RecallStatus()
{
	if(_FreeOrbitMFDState.m_Reference == NULL) return; //Static data not initialized

	m_Reference  = _FreeOrbitMFDState.m_Reference;
	m_Target     = _FreeOrbitMFDState.m_Target;
	m_ShowAlt    = _FreeOrbitMFDState.m_ShowAlt;
	m_Frame      = _FreeOrbitMFDState.m_Frame;
	m_Projection = _FreeOrbitMFDState.m_Projection;

	m_Mode       = _FreeOrbitMFDState.m_Mode;
}

// Return button labels
char *COrbitMFD::ButtonLabel(int bt)
{
	// The labels for the two buttons used by our MFD mode
	static char *label[9] =
	{"REF", "AR", "TGT", "NT", "MOD", "FRM", "PRJ", "DST", "HUD"};

	return (bt < 9 ? label[bt] : NULL);
}

// Return button menus
int COrbitMFD::ButtonMenu(const MFDBUTTONMENU **menu) const
{
	// The menu descriptions for the two buttons
	static const MFDBUTTONMENU mnu[9] = {
		{"Orbit reference",  NULL,              'R'},
		{"Auto reference",   NULL,              'A'},
		{"Select target",    NULL,              'T'},
		{"Unselect target",  NULL,              'N'},
		{"Display mode",     NULL,              'M'},
		{"Reference frame", "ECL/EQU",          'F'},
		{"Projection plane", NULL,              'P'},
		{"Distance display", "Radius/Altitude", 'D'},
		{"Copy data to HUD", NULL,              'H'}
	};
	if (menu) *menu = mnu;

	return 9; // return the number of buttons used
}

bool COrbitMFD::ConsumeButton(int bt, int event)
{
	static DWORD keys[9] =
	{OAPI_KEY_R, OAPI_KEY_A, OAPI_KEY_T, OAPI_KEY_N, OAPI_KEY_M, OAPI_KEY_F,
	OAPI_KEY_P, OAPI_KEY_D, OAPI_KEY_H};

	if(bt >= 9) return false;
	if(!(event & PANEL_MOUSE_LBDOWN)) return false;

	return ConsumeKeyBuffered(keys[bt]);
}

bool COrbitMFD::ConsumeKeyBuffered(DWORD key)
{
	switch(key)
	{
	case OAPI_KEY_R:
		btnREF();
		return true;
	case OAPI_KEY_A:
		btnAR();
		return true;
	case OAPI_KEY_T:
		btnTGT();
		return true;
	case OAPI_KEY_N:
		btnNT();
		return true;
	case OAPI_KEY_M:
		btnMOD();
		return true;
	case OAPI_KEY_F:
		btnFRM();
		return true;
	case OAPI_KEY_P:
		btnPRJ();
		return true;
	case OAPI_KEY_D:
		btnDST();
		return true;
	case OAPI_KEY_H:
		btnHUD();
		return true;
	}

	return false;
}

void COrbitMFD::btnREF()
{
	oapiCelBodyPopup(this, "Orbit MFD: Reference", refMenuResponse, 0);
}

void COrbitMFD::btnAR()
{
	autoReference();
	InvalidateDisplay();
}

void COrbitMFD::btnTGT()
{
	oapiTargetPopup(this, "Orbit MFD: Target", tgtMenuResponse, m_Reference, 0);
}

void COrbitMFD::btnNT()
{
	m_Target = NULL;
	if(m_Projection == ePTarget) m_Projection = ePFrame;
	InvalidateDisplay();
}

void COrbitMFD::btnMOD()
{
	m_Mode = eMode(int(m_Mode) + 1);
	if(m_Mode == eMMax) m_Mode = eMBoth;
	InvalidateDisplay();
}

void COrbitMFD::btnFRM()
{
	m_Frame = eFrame(int(m_Frame) + 1);
	if(m_Frame == eFMax) m_Frame = eFEcliptic;
	InvalidateDisplay();
}

void COrbitMFD::btnPRJ()
{
	m_Projection = eProjection(int(m_Projection) + 1);
	if(m_Projection == ePMax) m_Projection = ePFrame;
	if(m_Projection == ePTarget && m_Target == NULL) m_Projection = ePFrame;
	InvalidateDisplay();
}

void COrbitMFD::btnDST()
{
	m_ShowAlt = !m_ShowAlt;
	InvalidateDisplay();
}

void COrbitMFD::btnHUD()
{
	oapiSetHUDReference(m_Reference);
}

void COrbitMFD::autoReference()
{
	VESSELSTATUS status;
	m_Vessel->GetStatus(status);
	m_Reference = status.rbody;
}

void COrbitMFD::Update(HDC hDC)
{
	//Calculating orbital data
	calcDisplayData();

	//Title
	{
		char title[32];
		strcpy(title, "Orbit: ");
		oapiGetObjectName(m_Reference, title+7, 31-7);
		Title (hDC, title);
	}

	//Frame, Projection indicators
	{
		char *frm = "ECL";
		if(m_Frame == eFEquator) frm = "EQU";

		char *prj = frm;
		if(m_Projection == ePShip)   prj = "SHP";
		if(m_Projection == ePTarget) prj = "TGT";

		//White parts
		TextOut(hDC, int(0.718*m_Width) - 2, 1, frm, strlen(frm));
		TextOut(hDC, int(0.924*m_Width) - 2, 1, prj, strlen(prj));

		//Grey parts
		SetTextColor(hDC, MFD_GREY);
		TextOut(hDC, int(0.629*m_Width) - 2, 1, "Frm", 3);
		TextOut(hDC, int(0.830*m_Width) - 2, 1, "Prj", 3);
	}

	//G-field contribution indicator
	{
		bool makeRed = false;
		double Gpart = calcGContribution(makeRed);

		char Gpart_txt[16];
		sprintf(Gpart_txt, "G %.2f", Gpart);

		//Variable color:
		if(makeRed)
			{SetTextColor(hDC, OrbitMFDRedAlertColor);}
		else if(Gpart < 0.8)
			{SetTextColor(hDC, OrbitMFDYellowAlertColor);}
		else
			{SetTextColor(hDC, OrbitMFDGreenAlertColor);}

		TextOut(hDC, int(0.423*m_Width), m_Height-15, Gpart_txt, strlen(Gpart_txt));
	}

	//Graphical visualisation
	if(m_Mode == eMGraphics || m_Mode == eMBoth)
	{
		kostElements2Shape(&(m_VesselData.elements), &(m_VesselData.shape));
		projectData(m_VesselData);

		if(m_Target != NULL)
		{
			kostElements2Shape(&(m_TargetData.elements), &(m_TargetData.shape));
			projectData(m_TargetData);
		}

		calcScaleAndRange();

		//This allows SetDCPenColor to work:
		SelectObject(hDC, GetStockObject(DC_PEN));

		drawOrbitShape(hDC, OrbitMFDVesselOrbitColor, m_VesselData);

		if(m_Target != NULL)
			drawOrbitShape(hDC, OrbitMFDTargetOrbitColor, m_TargetData);

		//Planet surface:
		SetDCPenColor(hDC, OrbitMFDPlanetColor);
		int radius = int(m_Scale * oapiGetSize(m_Reference));
		if(radius > 0)
			Ellipse(hDC,
				m_Width/2 - radius, m_Height/2 - radius,
				m_Width/2 + radius, m_Height/2 + radius);
	}

	//Orbital element lists
	if(m_Mode == eMList || m_Mode == eMBoth)
	{
		m_VesselData.title = "OSC.EL.";
		SetTextColor(hDC, OrbitMFDVesselOrbitColor);
		drawDisplayData(hDC, 3, 16, m_VesselData);


		if(m_Target != NULL)
		{
			char title[16];
			m_TargetData.title = title;
			oapiGetObjectName(m_Target, title, 15);

			SetTextColor(hDC, OrbitMFDTargetOrbitColor);
			drawDisplayData(hDC, 3 + int(0.69*m_Width), 16, m_TargetData);
		}
	}
}

void COrbitMFD::calcScaleAndRange()
{
	//Some points that should be inside the view:
	std::vector<kostVector3> points;

	//We're looking at 2D coordinates only, so
	//3D relationships like |PeD| <= |Dst| can not
	//be used.

	points.push_back(m_VesselData.state.pos);
	points.push_back(m_VesselData.shape.pe);
	if(m_VesselData.elements.e < 1.0)
		points.push_back(m_VesselData.shape.ap);

	if(m_Target != NULL)
	{
		points.push_back(m_TargetData.state.pos);
		points.push_back(m_TargetData.shape.pe);
		if(m_TargetData.elements.e < 1.0)
			points.push_back(m_TargetData.shape.ap);
	}

	double largest = 0.0;
	for(unsigned int i=0; i < points.size(); i++)
	{
		kostVector3 &vec = points[i];
		double len_2d = vec.x*vec.x + vec.y*vec.y;
		if(len_2d > largest) largest = len_2d;
	}
	largest = sqrt(largest); //Turn length^2 into length

	//Last 'point': size of the planet
	double rad = oapiGetSize(m_Reference);
	if(rad > largest) largest = rad;

	m_Scale = 0.45 * double(m_Width) / largest;
}

void COrbitMFD::calcDisplayData()
{
	//EQU frame orientation
	MATRIX3 refOrientation;
	oapiGetPlanetObliquityMatrix(m_Reference, &refOrientation);
	makeTranspose(refOrientation);

	kostReal refMu =
		KOST_GRAVITATIONAL_CONSTANT * (oapiGetMass(m_Reference) + m_Vessel->GetMass());

	m_Vessel->GetRelativePos(m_Reference, m_VesselData.state.pos);
	m_Vessel->GetRelativeVel(m_Reference, m_VesselData.state.vel);

	//To the frame coordinate system:
	if(m_Frame == eFEquator)
	{
		m_VesselData.state.pos = refOrientation * m_VesselData.state.pos;
		m_VesselData.state.vel = refOrientation * m_VesselData.state.vel;
	}

	/*
	Transform to standard right-handed coordinate system:
	y -> z
	z -> y
	*/
	std::swap<kostReal>(m_VesselData.state.pos.y, m_VesselData.state.pos.z);
	std::swap<kostReal>(m_VesselData.state.vel.y, m_VesselData.state.vel.z);

	//Calculate orbital elements
	kostStateVector2Elements(refMu,
		&(m_VesselData.state), &(m_VesselData.elements), &(m_VesselData.params));

	if(m_Target != NULL)
	{
		refMu =
			KOST_GRAVITATIONAL_CONSTANT * (oapiGetMass(m_Reference) + oapiGetMass(m_Target));

		oapiGetRelativePos(m_Target, m_Reference, &(m_TargetData.state.pos));
		oapiGetRelativeVel(m_Target, m_Reference, &(m_TargetData.state.vel));

		//To the frame coordinate system:
		if(m_Frame == eFEquator)
		{
			m_TargetData.state.pos = refOrientation * m_TargetData.state.pos;
			m_TargetData.state.vel = refOrientation * m_TargetData.state.vel;
		}

		/*
		Transform to standard right-handed coordinate system:
		y -> z
		z -> y
		*/
		std::swap<kostReal>(m_TargetData.state.pos.y, m_TargetData.state.pos.z);
		std::swap<kostReal>(m_TargetData.state.vel.y, m_TargetData.state.vel.z);

		//Calculate orbital elements
		kostStateVector2Elements(refMu,
			&(m_TargetData.state), &(m_TargetData.elements), &(m_TargetData.params));
	}
}

void COrbitMFD::projectData(SDisplayData &data)
{
	if(m_Projection == ePFrame) return; //Already good -> nothing to do

	kostElements projElems;

	switch(m_Projection)
	{
	default:
		return;
	case ePShip:
		projElems = m_VesselData.elements;
		break;
	case ePTarget:
		projElems = m_TargetData.elements;
		break;
	}

	kostMatrix3 rmat; //Rotation matrix

	kostMatrix3 LAN, Inc;
	kostMakeZRotm(&LAN, -projElems.theta);
	kostMakeXRotm(&Inc, -projElems.i);

	rmat = Inc * LAN;

	kostMakeTransposem(&LAN);
	rmat = LAN * rmat;

	data.state.pos *= rmat;
	data.state.vel *= rmat;
	data.shape.pe  *= rmat;
	data.shape.ap  *= rmat;
	data.shape.dn  *= rmat;
	data.shape.an  *= rmat;
	for(unsigned int i=0; i < data.shape.numPoints; i++)
		data.shape.points[i] *= rmat;
}

double COrbitMFD::calcGContribution(bool &makeRed) const
{
	//Old method, proposed by Hielor:
	/*
	VECTOR3 Gtot_vec;
	m_Vessel->GetWeightVector(Gtot_vec);

	double Gtot = kostAbsv(&Gtot_vec);
	double Gref =
		KOST_GRAVITATIONAL_CONSTANT *
		m_Vessel->GetMass() * oapiGetMass(m_Reference) / 
		kostAbs2v(&m_VesselData.state.pos);

	double ret = Gref / Gtot;
	makeRed = ret < 0.5;
	return ret;
	*/

#ifdef ORBIT_MFD_G_INDICATOR_COMPATIBLE

	//How it's probably done by Orbiter:
	makeRed = false;
	double Gref = oapiGetMass(m_Reference) / kostAbs2v(&m_VesselData.state.pos);

	double Gtot = 0.0; //in (m/s^2) / G, NOT in Newton!

	unsigned int numGbodies = oapiGetGbodyCount();
	for(unsigned int i=0; i < numGbodies; i++)
	{
		OBJHANDLE obj = oapiGetGbodyByIndex(i);

		VECTOR3 pos;
		m_Vessel->GetRelativePos(obj, pos);

		double Gobj = oapiGetMass(obj) / kostAbs2v(&pos);
		if(Gobj > Gref && obj != m_Reference) makeRed = true;

		Gtot += Gobj;
	}

	return Gref / Gtot;

#endif


#ifdef ORBIT_MFD_G_INDICATOR_CJP
	//Brand new method, based on relative acceleration (a_vessel - a_reference):
	/*
	The idea is to compare the acceleration caused by the reference
	body with the DIFFERENCE between the vessel and the reference of
	the acceleration caused by other bodies.

	If a disturbing body causes (almost) the same acceleration on
	reference and body, then the relative motion isn't affected(*).
	So, even when this disturbing force is greater than the force
	between reference and vessel, the orbit between reference and
	vessel can still be stable.

	The current method doesn't seem to work the way I want it.
	For instance, at G 0.50 , an earth orbit seems to be
	very unstable. When leaving earth, the orbit seems to
	change already around G 0.8.

	(*) Strictly speaking, the relative motion is measured in an
	accelerating reference frame, which is non-inertial. It is not
	yet investigated whether this is a problem in actual situations.

	All accelerations are in (m/s^2) / G
	*/
	kostVector3 a_rel = kostConstructv(0.0, 0.0, 0.0);

	unsigned int numGbodies = oapiGetGbodyCount();
	for(unsigned int i=0; i < numGbodies; i++)
	{
		OBJHANDLE obj = oapiGetGbodyByIndex(i);
		if(obj == m_Reference) continue; //only count disturbing bodies

		double m_obj = oapiGetMass(obj);

		kostVector3 pos, dir;

		//Vessel acceleration by body:
		m_Vessel->GetRelativePos(obj, pos);
		dir = kostNormalv(&pos);
		kostVector3 a_vessel = dir * (m_obj / kostAbs2v(&pos));

		//Reference acceleration by body:
		oapiGetRelativePos(m_Reference, obj, &pos);
		dir = kostNormalv(&pos);
		kostVector3 a_reference = dir * (m_obj / kostAbs2v(&pos));

		a_rel += a_vessel - a_reference;
	}

	kostVector3 dir = kostNormalv(&m_VesselData.state.pos);
	kostVector3 a_ref = dir * (oapiGetMass(m_Reference) / kostAbs2v(&m_VesselData.state.pos));

	double a_ref_abs = kostAbsv(&a_ref);
	double a_rel_abs = kostAbsv(&a_rel);

	//Use scalar addition to ensure ret <= 1.0
	double a_tot = a_rel_abs + a_ref_abs;

	kostReal ret = a_ref_abs / a_tot;
	makeRed = ret < 0.5;
	return ret;
#endif
}

void COrbitMFD::drawDisplayData(HDC hDC, int x, int y, const SDisplayData &data)
{
	char *str[] =
	{
		"-----------",
		"SMa ",
		"SMi ",
		"PeR ",
		"ApR ",
		"Rad ",
		"Ecc ",
		"T   ",
		"PeT ",
		"ApT ",
		"Vel ",
		"Inc ",
		"LAN ",
		"LPe ",
		"AgP ",
		"TrA ",
		"TrL ",
		"MnA ",
		"MnL "
	};

	char lines[19][64];

	unsigned int i;
	for(i=0; i < 19; i++)
		strcpy(lines[i], str[i]);

	//Title:
	{
		unsigned int len = strlen(data.title);
		unsigned int linelen = strlen(lines[0]);
		if(len >= linelen)
		{
			for(unsigned int i=0; i < linelen; i++)
				lines[0][i] = data.title[i];
		}
		else
		{
			unsigned int offset = (linelen - len) / 2; //rounded downwards
			for(unsigned int i=0; i < len; i++)
				lines[0][offset+i] = data.title[i];
		}
	}

	formatValue(lines[ 1] + 3,            data.elements.a); //SMa
	formatValue(lines[ 2] + 3,            data.params.SMi); //SMi
	formatValue(lines[ 3] + 3,            data.params.PeD); //PeR
	formatValue(lines[ 4] + 3,            data.params.ApD); //ApR
	formatValue(lines[ 5] + 3, kostAbsv(&data.state.pos )); //Rad
	sprintf    (lines[ 6] + 3, "% 0.*f",
		  1+numDecimals(data.elements.e), data.elements.e); //Ecc
	formatValue(lines[ 7] + 3,            data.params.T  ); //T
	formatValue(lines[ 8] + 3,            data.params.PeT); //PeT
	formatValue(lines[ 9] + 3,            data.params.ApT); //ApT
	formatValue(lines[10] + 3, kostAbsv(&data.state.vel )); //Vel

	kostReal Inc_deg = data.elements.i      * (180.0/M_PI);
	sprintf    (lines[11] + 3, "% 7.2f",         Inc_deg); //Inc

	kostReal LAN_deg = data.elements.theta  * (180.0/M_PI);
	sprintf    (lines[12] + 3, "% 7.2f",         LAN_deg); //LAN

	kostReal LPe_deg = data.elements.omegab * (180.0/M_PI);
	sprintf    (lines[13] + 3, "% 7.2f",         LPe_deg); //LPe

	kostReal AgP_deg = data.params.AgP * (180.0/M_PI);
	sprintf    (lines[14] + 3, "% 7.2f",         AgP_deg); //AgP

	kostReal TrA_deg = data.params.TrA * (180.0/M_PI);
	sprintf    (lines[15] + 3, "% 7.2f",         TrA_deg); //TrA

	kostReal TrL_deg = data.params.TrL * (180.0/M_PI);
	sprintf    (lines[16] + 3, "% 7.2f",         TrL_deg); //TrL

	kostReal MnA_deg = data.params.MnA * (180.0/M_PI);
	sprintf    (lines[17] + 3, "% 7.2f",         MnA_deg); //MnA

	kostReal MnL_deg = data.elements.L      * (180.0/M_PI);
	sprintf    (lines[18] + 3, "% 7.2f",         MnL_deg); //MnL


	//Convert to Alt:
	if(m_ShowAlt)
	{
		strcpy(lines[3], "PeA ");
		strcpy(lines[4], "ApA ");
		strcpy(lines[5], "Alt ");

		kostReal refR = oapiGetSize(m_Reference);
		formatValue(lines[3] + 3,            data.params.PeD - refR); //PeA
		formatValue(lines[4] + 3,            data.params.ApD - refR); //ApA
		formatValue(lines[5] + 3, kostAbsv(&data.state.pos ) - refR); //Alt
	}

	//Some data is not available for hyperbola:
	if(data.elements.e >= 1.0)
	{
		strcpy(lines[ 4] + 3, " N/A"); //ApR
		strcpy(lines[ 7] + 3, " N/A"); //T
		strcpy(lines[ 9] + 3, " N/A"); //ApT
		strcpy(lines[18] + 3, " N/A"); //MnL
	}

	int txtPos = y;
	int txtIncr = (m_Height - 10) / 22;

	SetTextAlign(hDC,TA_LEFT);

	for(i=0; i < 19; i++)
	{
		TextOut(hDC, x, txtPos, lines[i], strlen(lines[i]));
		txtPos += txtIncr;
	}
}

void COrbitMFD::drawOrbitShape(HDC hDC, COLORREF color, const SDisplayData &data)
{
	HBRUSH fillBrush = CreateSolidBrush(color);
	SetDCPenColor(hDC, color);

	kostVector3 pt;

	//Orbit shape:
	POINT *points = new POINT[data.shape.numPoints];
	for(unsigned int i=0; i < data.shape.numPoints; i++)
	{
		pt = data.shape.points[i] * m_Scale;
		points[i].x = m_Width/2 + int(pt.x);
		points[i].y = m_Height/2 - int(pt.y); //Flipped y-axis
	}
	Polyline(hDC, points, data.shape.numPoints);
	delete [] points;

	//Object line:
	pt = data.state.pos * m_Scale;
	MoveToEx(hDC,  m_Width/2,                 m_Height/2, NULL);
	LineTo(  hDC,  m_Width/2 + int(pt.x), m_Height/2 - int(pt.y));

	bool hasDN = kostAbs2v(&data.shape.dn) > 0.1;
	bool hasAN = kostAbs2v(&data.shape.an) > 0.1;

	//Pe
	{
		pt = data.shape.pe * m_Scale;
		int x = m_Width/2 + int(pt.x);
		int y = m_Height/2 - int(pt.y); //Flipped y-axis

		HBRUSH oldBrush = (HBRUSH)SelectObject(hDC, fillBrush);
		Ellipse(hDC, x-3, y-3, x+4, y+4);
		SelectObject(hDC, oldBrush);
	}

	//Ap
	if(data.elements.e < 1.0)
	{
		pt = data.shape.ap * m_Scale;
		int x = m_Width/2 + int(pt.x);
		int y = m_Height/2 - int(pt.y); //Flipped y-axis
		Ellipse(hDC, x-3, y-3, x+4, y+4);
	}

	//DN
	if(hasDN)
	{
		pt = data.shape.dn * m_Scale;
		int x = m_Width/2 + int(pt.x);
		int y = m_Height/2 - int(pt.y); //Flipped y-axis
		Rectangle(hDC, x-3, y-3, x+4, y+4);
	}

	//AN
	if(hasAN)
	{
		pt = data.shape.an * m_Scale;
		int x = m_Width/2 + int(pt.x);
		int y = m_Height/2 - int(pt.y); //Flipped y-axis
		HBRUSH oldBrush = (HBRUSH)SelectObject(hDC, fillBrush);
		Rectangle(hDC, x-3, y-3, x+4, y+4);
		SelectObject(hDC, oldBrush);
	}

	//Line between DN and AN
	if(hasDN && hasAN)
	{
		//Dashed pen:
		HPEN pen = CreatePen(PS_DOT, 1, color);
		HPEN oldpen = (HPEN)SelectObject(hDC, pen);

		pt = data.shape.dn * m_Scale;
		MoveToEx(hDC,  m_Width/2 + int(pt.x), m_Height/2 - int(pt.y), NULL);

		pt = data.shape.an * m_Scale;
		LineTo(  hDC,  m_Width/2 + int(pt.x), m_Height/2 - int(pt.y));

		SelectObject(hDC, oldpen);
		DeleteObject(pen);
	}

	DeleteObject(fillBrush);
}


